/**********************************************************************
 * mod_dpmput: implements the PUT and DELETE methods on a dpm https server
 * Author: Andreas-Joachim Peters CERN(2007)
 *
 * usage:
 *	EnableDpmPut	On|Off
 *		default value:	Off
 *		context:	Directory
 *		effect:		enables (or disables) the PUT method for a
 *				whole directory.
 *
 *	EnableDpmDelete	On|Off
 *		default value:	Off
 *		context:	Directory
 *		effect:		enables (or disables) the DELETE method for a
 *				whole directory.
 *	umask		<octal value>
 *		default value:	007
 *		context:	Directory
 *		effect:		sets the umask for a whole directory (see umask(1)).
 */
/*
 Copyright (c) Members of the EGEE Collaboration. 2004.
 See http://www.eu-egee.org/partners/ for details on the copyright holders.
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
 
     http://www.apache.org/licenses/LICENSE-2.0
 
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
*/

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "apr_strings.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#define DEF_UMASK (S_IROTH | S_IWOTH | S_IXOTH)

typedef struct
{
  int put_enabled;
  int delete_enabled;
  int umask;
} put_config_rec;

typedef struct {
  char   path[4096];      /* lfn */
  char   sfn[4096];       /* sfn */
  char   ip[64];          /* the clients IP address */
  char   method[64];      /* http method e.g. GET/POST */
  char   keyhash[9];      /* hash of the public key to be used (not used yet) */
  time_t exptime;         /* time when the authorization expires */
  char   clientid[4096];  /* ID of the client = DN */
  char   token[8192];     /* the full token */
  char   signature[8192]; /* signature for 'token' */
  char   redirectorhost[4096]; /* hostname of the DPM redirector node */
  char   r_token[4096];   /* request token as used in dpm_get & dpm_put */
} authz_info;

module AP_MODULE_DECLARE_DATA dpmput_module;

#define BLOCKSIZE 2048

const char *put_set_umask (cmd_parms *cmd, void* cfg, const char * arg)
{
  ((put_config_rec *)cfg)->umask=strtol(arg,(char **)NULL,8);
  return NULL;
}

static const command_rec put_cmds[] = {
	AP_INIT_FLAG("EnableDpmPut",    ap_set_flag_slot, (void *)APR_OFFSETOF(put_config_rec, put_enabled), OR_AUTHCFG, "Limited to 'on' or 'off'"),
	AP_INIT_FLAG("EnableDpmDelete", ap_set_flag_slot, (void *)APR_OFFSETOF(put_config_rec, delete_enabled), OR_AUTHCFG, "Limited to 'on' or 'off'"), 
	AP_INIT_TAKE1("umask",       put_set_umask,    (void *)APR_OFFSETOF(put_config_rec, umask), OR_AUTHCFG,  "numeric umask") ,
	{ NULL }
};

static void *create_put_dir_config (apr_pool_t *p, char *d)
{
  put_config_rec * sec = (put_config_rec *)apr_pcalloc (p, sizeof(put_config_rec));

  if (!sec) return NULL; /* no memory... */

  sec -> put_enabled	=0;
  sec -> delete_enabled	=0;
  sec -> umask          =DEF_UMASK;
  return sec;
}

void make_dirs(apr_pool_t * p, const char * filename, int umask)
{
  char* sto = apr_pstrdup(p, filename);
  char* slash=sto;
  char* dirname=sto;
  while(slash=strchr(slash+1,'/'))
  {
    *slash='\0';
    *dirname='/';
    mkdir(sto,(S_IRWXU | S_IRWXG | S_IRWXO) & ~umask);
    dirname=slash;
  }
}


/**********************************************************************
 * function to parse a query string and fill the authz_info structure
 */

static authz_info* get_authz_info(const char* query) {
  char* tokenize=0;
  char* keyval=0;
  char *search="&";
  char *token=0;
  char *keyend=0;
  char  lkey[8192];
  char  value[8192];
  char *tokenptr;
  char *httpstoken=0;
  char *httpsauthz=0;
  authz_info* authz=0;

  
  if (query) {
    authz = (authz_info*) malloc(sizeof(authz_info));
    memset(authz,0, sizeof(authz_info));
    /* extract httpstoken & httpsauthz */
    tokenize = strdup(query);
    tokenptr = tokenize;
    while ( (token=strtok(tokenptr, search))) {
      keyend = strchr(token,'=');
      if (!keyend) {
	tokenptr=NULL;
	continue;
      }
      
      strncpy(lkey,token,keyend-token );
      lkey[keyend-token]=0;
      strncpy(value,keyend+1,strlen(token)-(keyend-token+1));
      value[strlen(token)-(keyend-token+1)]=0;
      tokenptr=NULL;
      
      if (!strcmp(lkey,"httpstoken")) {
	httpstoken = strdup(value);
      }
      
      if (!strcmp(lkey,"httpsauthz")) {
	httpsauthz = strdup(value);
      }
    }
    free(tokenize);
  } else {
    return 0;
  }
  
  /* if httpstoken was given, parse it ... */
  if (httpstoken) {
    /* tokens are made as <path>@<client-ip>:<sfn>:<key-hash>:<expirationtime>:<client-id>:<redirector-host>:<r-token> */
    tokenize = strdup(httpstoken);
    if ((token = strtok(tokenize, "@"))) { 
      strcpy(authz->path,token);   
      if ((token = strtok(NULL, ":"))) {
	strcpy(authz->ip,token);
	if ((token = strtok(NULL, ":"))) {
	  strcpy(authz->method,token);
	  if ((token = strtok(NULL, ":"))) {
	    strcpy(authz->sfn,token);	    
	    if ((token = strtok(NULL, ":"))) {
	      strcpy(authz->keyhash,token);	      
	      if ((token = strtok(NULL, ":"))) {
		authz->exptime = (time_t) atol(token);
		if ((token = strtok(NULL, ":"))) {
		  strcpy(authz->clientid,token);
		  if ((token = strtok(NULL, ":"))) {
		    strcpy(authz->redirectorhost,token);
		    if ((token = strtok(NULL, ":"))) {
		      strcpy(authz->r_token,token);
		    }
		  }
		} 
	      }
	    }
	  }
	}
      }
    }
  }

  if (httpstoken) {
    strcpy(authz->token,httpstoken);
    /* cert DNs come f.e. with escaped spaces  */
    ap_unescape_url(authz->token);
  }

  if (httpsauthz) {
    strcpy(authz->signature,httpsauthz);
  }

  return authz;
}

int do_put(request_rec *r, int ispost)
{
  put_config_rec *sec = (put_config_rec *)ap_get_module_config (r->per_dir_config, &dpmput_module);
  apr_file_t* f;
  apr_status_t f_status;
  int result=OK, len=0;
  mode_t old_umask;
  char * buffer;
  int nblocks=0;
  char *content;
  char *stream;
  int offset=0;
  char *coffset;
  apr_off_t totalbytes=0;
  char redirect[4096];

  authz_info* authz;
  authz=get_authz_info(r->args);

  if(!sec->put_enabled) {
    return DECLINED;
  }

  if (!authz) {
    return DECLINED;
  }
  
  if(r->path_info)	/* a directory did not exist */
  {
    r->filename = apr_pstrcat(r->pool, r->filename, r->path_info, NULL);
    make_dirs(r->pool,r->filename,sec->umask);
  }

  f_status = apr_file_open(&f, r->filename,APR_READ|APR_WRITE | APR_CREATE | APR_TRUNCATE, APR_UREAD | APR_UWRITE | APR_GREAD |APR_WREAD,r->pool);
  if (f_status != APR_SUCCESS)
  {
    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
		  "file permissions deny server write access");
    return DECLINED;
  }


  if((result=ap_setup_client_block(r,REQUEST_CHUNKED_DECHUNK))==OK)
    {
    if(ap_should_client_block(r))
    {
      buffer = apr_palloc(r->pool, BLOCKSIZE+1);
      buffer[BLOCKSIZE]=0;
      while((len=ap_get_client_block(r,buffer,BLOCKSIZE))>0) {
	nblocks++;
	offset=0;
	if (ispost) {
	  if (nblocks==1) {
	    if ((content=strstr(buffer,"Content-Type:"))) {
	      coffset = strstr(content,"\r\n\r\n");
	      if (coffset) {
		coffset += 4;
		offset = ( (int) coffset - (int) buffer);;
	      } else {
		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
			      "post request does not contain stream type");
		return DECLINED;
	      }
	    } else {
	      ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
			    "post request does not contain Content-type");
	      return DECLINED;
	    }
	  }
	}

	apr_size_t alen=(len-offset);
	f_status= apr_file_write(f, buffer+offset,&alen);
	if (f_status != APR_SUCCESS) {
	  ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
			"error during file write");
	  return DECLINED;
	} else {
	  totalbytes += alen;
	}
      }
    }
    if (ispost) {
      char endbuf[128];
      char* eraseptr;
      apr_off_t seekoffset=-128;
      apr_size_t endblocksize=128;
      sprintf(endbuf,"");
      f_status = apr_file_seek(f,APR_END,&seekoffset);
      if (f_status != APR_SUCCESS) {
	ap_log_rerror(APLOG_MARK,APLOG_ERR, 0, r,
		      "error during file seek");
	return DECLINED;
      }

      f_status = apr_file_read(f,endbuf,&endblocksize);
      if (f_status != APR_SUCCESS) {
	ap_log_rerror(APLOG_MARK,APLOG_ERR, 0, r,
		      "error during file read");
	return DECLINED;
      }

      if ( (eraseptr = strstr(endbuf,"\r\n----------------"))) {
	f_status = apr_file_trunc(f, totalbytes-128+(eraseptr-endbuf));
	if (f_status != APR_SUCCESS) {
	  ap_log_rerror(APLOG_MARK,APLOG_ERR, 0, r,
			"error during file truncate");
	  return DECLINED;
	}
	totalbytes = totalbytes-128+(eraseptr-endbuf);

      } else {
	if (f_status != APR_SUCCESS) {
	  ap_log_rerror(APLOG_MARK,APLOG_ERR, 0, r,
			"error in form format - cannot find form end!");
	  return DECLINED;      
	}
      }
    }
  }

  apr_file_close(f);
  chmod(r->filename, (S_IRUSR |S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH) & ~sec->umask);

  // semd the redirection to the client
  sprintf(redirect,"https://%s%s?metacmd=putdone&dpmtoken=%s\n",authz->redirectorhost,authz->path,authz->r_token);

  if (ispost) {
    apr_table_setn(r->headers_out, "Location", redirect);
    return HTTP_MOVED_TEMPORARILY;
  } else {
    return OK;
  }
}

int do_delete(request_rec *r)
{
  put_config_rec *sec = (put_config_rec *)ap_get_module_config (r->per_dir_config, &dpmput_module);

  if(!sec->delete_enabled) return DECLINED;
  if(r->finfo.protection == 0) return DECLINED;
  if(unlink(r->filename))
  {
    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
		  "file permissions deny file deletion");
    return DECLINED;
  }

  return OK;
}

int do_post(request_rec *r) 
{
  do_put(r,1);
}

int put_handler(request_rec *r)
{
  if(r->proxyreq) return DECLINED;
  if(r->method_number == M_PUT)
  {
#ifdef CHARSET_EBCDIC
	  ap_checkconv(r);
#endif
	  return do_put(r,0);
  }
  else
  if(r->method_number == M_DELETE) return do_delete(r);
  else
  if(r->method_number == M_POST) return do_post(r);

  return DECLINED;
}

/**********************************************************************
 * register put hooks in Apache
 */

 static void mod_put_register_hooks(apr_pool_t *p)
{
  ap_hook_handler(put_handler, NULL, NULL, APR_HOOK_LAST);
}


module AP_MODULE_DECLARE_DATA dpmput_module = {
   STANDARD20_MODULE_STUFF,
   create_put_dir_config,	/* dir config creater */
   NULL,			/* dir merger --- default is to override */
   NULL,			/* server config */
   NULL,			/* merge server config */
   put_cmds,			/* command table */
   mod_put_register_hooks,      /* callback for registering hooks */
};
